@@ -29,9 +29,23 @@ XML::Stream - Creates and XML Stream connection and parses return data
 
 =head1 METHODS
 
-  Connect(name=>string,      - opens a tcp connection to the specified
+  new(debug=>string,       - creates the XML::Stream object.  debug should
+      debugfh=>FileHandle)   be set to the path for the debug log to be
+                             written.  If set to "stdout" then the debug
+                             will go there.   Also, you can specify a 
+                             filehandle that already exists and use that.
+
+  Connect(hostname=>string,  - opens a tcp connection to the specified
           port=>integer,       server and sends the proper opening XML
-          namespace=>string)   Stream tag.  All fields are required.
+          myhostname=>string,  Stream tag.  hostname, port, and namespace
+          namespace=>array,    are required.  namespaces allows you
+          namespaces=>array)   to use XML::Stream::Namespace objects.
+                               myhostname should not be needed but if 
+                               the module cannot determine your hostname 
+                               properly (check the debug log), set this 
+                               to the correct value, or if you want
+                               the other side of the stream to think that
+                               you are someone else.
 
   Disconnect() - sends the proper closing XML tag and closes the socket
                  down.
@@ -39,7 +53,7 @@ XML::Stream - Creates and XML Stream connection and parses return data
   Process(integer) - waits for data to be available on the socket.  If 
                      a timeout is specified then the Process function
                      waits that period of time before returning nothing.  
-                     If a timeout period is not specified then the 
+                     If a timeout period is not specified then the
                      function blocks until data is received.
 
   OnNode(function pointer) - sets the callback used to handle the
@@ -55,6 +69,10 @@ XML::Stream - Creates and XML Stream connection and parses return data
                  does no checking if valid XML was sent or not.  Best
                  behavior when sending information.
 
+  GetErrorCode() - returns a string that will hopefully contain some
+                   useful information about why Process or Connect
+                   returned an undef to you.
+
 =head1 EXAMPLES
 
   ##########################
@@ -64,12 +82,17 @@ XML::Stream - Creates and XML Stream connection and parses return data
 
   $stream = new XML::Stream;
 
-  $stream->Connect(name => "jabber.org", 
-                   port => 5222, 
-                   namespace => "jabber:client") || die $!;
+  my $status = $stream->Connect(hostname => "jabber.org", 
+                                port => 5222, 
+                                namespace => "jabber:client");
 
-  while($node = $stream->Process())
-  {
+  if (!defined($status)) {
+    print "ERROR: Could not connect to server\n";
+    print "       (",$stream->GetErrorCode(),")\n";
+    exit(0);
+  }
+
+  while($node = $stream->Process()) {
     # do something with $node
   }
 
@@ -83,19 +106,17 @@ XML::Stream - Creates and XML Stream connection and parses return data
 
   $stream = new XML::Stream;
   $stream->OnNode(\&noder);
-  $stream->Connect(name => "jabber.org",
+  $stream->Connect(hostname => "jabber.org",
 		   port => 5222,
 		   namespace => "jabber:client",
 		   timeout => undef) || die $!;
 
   # Blocks here forever, noder is called for incoming 
   # packets when they arrive.
-  while(1) {
-    $stream->Process();
-  }
-
-  $stream->Disconnect();
+  while(defined($stream->Process())) { }
 
+  print "ERROR: Stream died (",$stream->GetErrorCode(),")\n";
+  
   sub noder
   {
     my $node = shift;
@@ -105,6 +126,7 @@ XML::Stream - Creates and XML Stream connection and parses return data
 =head1 AUTHOR
 
 Tweaked, tuned, and brightness changes by Ryan Eatmon, reatmon@ti.com
+in May of 2000.
 Colorized, and Dolby Surround sound added by Thomas Charron,
 tcharron@jabber.org
 By Jeremie in October of 1999 for http://etherx.jabber.org/streams/
@@ -118,13 +140,15 @@ it under the same terms as Perl itself.
 
 require 5.003;
 use strict;
-use Carp;
+use Socket;
+use Sys::Hostname;
 use IO::Socket;
 use IO::Select;
 use XML::Parser;
-use vars qw($VERSION);
+use vars qw($VERSION $STREAMERROR);
 
-$VERSION = "0.1";
+$VERSION = "1.0";
+$STREAMERROR = "";
 
 use XML::Stream::Namespace;
 ($XML::Stream::Namespace::VERSION < $VERSION) &&
@@ -133,14 +157,62 @@ use XML::Stream::Namespace;
 sub new {
   my $self = { };
 
+  bless($self);
+
+  my %args;
+  while($#_ >= 0) { $args{ lc pop(@_) } = pop(@_); }
+
+  if (exists($args{debugfh}) && ($args{debugfh} ne "")) {
+    $self->{DEBUGFILE} = $args{debugfh};
+    $self->{DEBUG} = 1;
+  }
+  if ((exists($args{debugfh}) && ($args{debugfh} eq "")) && 
+       (exists($args{debug}) && ($args{debug} ne ""))) {
+    $self->{DEBUG} = 1;
+    if (lc($args{debug}) eq "stdout") {
+      open(DEBUG, ">STDOUT");
+      $self->{DEBUGFILE} = \*DEBUG;
+    } else {
+      if (-e $args{debug}) {
+	if (-w $args{debug}) {
+	  open(DEBUG, ">$args{debug}");
+	  $self->{DEBUGFILE} = \*DEBUG;
+	} else {
+	  print "WARNING: debug file ($args{debug}) is not writable by you\n";
+	  print "         No debug information being saved.\n";
+	  $self->{DEBUG} = 0;
+	}
+      } else {
+	if (open(DEBUG, ">$args{debug}")) {
+	  $self->{DEBUGFILE} = \*DEBUG;
+	} else {
+	  print "WARNING: debug file ($args{debug}) does not exist \n";
+	  print "         and is not writable by you.\n";
+	  print "         No debug information being saved.\n";
+	  $self->{DEBUG} = 0;
+	}
+      }
+    }
+  }
+
+  my $hostname = hostname();
+  my $address = gethostbyname($hostname) || 
+    die("Cannot resolve $hostname: $!");
+  my $fullname = gethostbyaddr($address,AF_INET) || 
+    die("Cannot re-resolve $hostname: $!");
+
+  $self->debug("XML::Stream: new: hostname = ($fullname)");
+
   #---------------------------------------------------------------------------
   # Setup the defaults that the module will work with.
   #---------------------------------------------------------------------------
-  $self->{SERVER} = {name => "",
+  $self->{SERVER} = {hostname => "",
 		     port => "", 
 		     sock => 0, 
-		     namespace => "", 
-		     timeout => 0};
+		     namespace => "",
+		     myhostname => $fullname,
+		     derivedhostname => $fullname,
+		     id => ""};
   
   #---------------------------------------------------------------------------
   # We are only going to use one callback, let the user call other callbacks
@@ -162,11 +234,23 @@ sub new {
   #---------------------------------------------------------------------------
   $self->{NODES} = ();
 
-  bless($self);
   return $self;
 }
 
 
+###########################################################################
+#
+# debug - prints the arguments to the debug log if debug is turned on.
+#
+###########################################################################
+sub debug {
+  my $self = shift;
+  return if !($self->{DEBUG});
+  my $fh = $self->{DEBUGFILE};
+  print $fh "@_\n";
+}
+
+
 ##############################################################################
 #
 # Connect - starts the stream by connecting to the server, sending the opening
@@ -177,23 +261,27 @@ sub new {
 ##############################################################################
 sub Connect {
   my $self = shift;
+  my $timeout = exists $_{timeout} ? delete $_{timeout} : "";
   while($#_ >= 0) { $self->{SERVER}{ lc pop(@_) } = pop(@_); }
 
   #---------------------------------------------------------------------------
   # Check some things that we have to know in order get the connection up
-  # and running.  Server name, port number, namespace, etc...
+  # and running.  Server hostname, port number, namespace, etc...
   #---------------------------------------------------------------------------
-  if ($self->{SERVER}{name} eq "") { 
-    $! = "Server name not specified";
-    return undef;
+  if ($self->{SERVER}{hostname} eq "") {
+    $self->SetErrorCode("Server hostname not specified");
+    return;
   }
   if ($self->{SERVER}{port} eq "") {
-    $! = "Server port not specified";
-    return undef;
+    $self->SetErrorCode("Server port not specified");
+    return;
   }
   if ($self->{SERVER}{namespace} eq "") {
-    $! = "Namespace not specified";
-    return undef;
+    $self->SetErrorCode("Namespace not specified");
+    return;
+  }
+  if ($self->{SERVER}{myhostname} eq "") {
+    $self->{SERVER}{myhostname} = $self->{SERVER}{derivedhostname};
   }
   
   #---------------------------------------------------------------------------
@@ -201,26 +289,34 @@ sub Connect {
   # abort ourselves and let the user check $! on his own.
   #---------------------------------------------------------------------------
   $self->{SERVER}{sock} = 
-    new IO::Socket::INET(PeerAddr => $self->{SERVER}{name}, 
+    new IO::Socket::INET(PeerAddr => $self->{SERVER}{hostname}, 
 			 PeerPort => $self->{SERVER}{port}, 
 			 Proto => 'tcp');
-  return undef unless $self->{SERVER}{sock};
+  return unless $self->{SERVER}{sock};
   $self->{SERVER}{sock}->autoflush(1);
   
-
+  #---------------------------------------------------------------------------
+  # Next, we build the opening handshake.
+  #---------------------------------------------------------------------------
+  my $stream = '<?xml version="1.0"?>';
+  $stream .= '<stream:stream ';
+  $stream .= 'xmlns:stream="http://etherx.jabber.org/streams" ';
+  $stream .= 'to="'.$self->{SERVER}{hostname}.'" ';
+  $stream .= 'from="'.$self->{SERVER}{myhostname}.'" ' if ($self->{SERVER}{myhostname} ne "");
+  $stream .= 'xmlns="'.$self->{SERVER}{namespace}.'" ';
+  $stream .= 'id="'.$self->{SERVER}{id}.'"' if (exists($self->{SERVER}{id}) && ($self->{SERVER}{id} ne ""));
   my $namespaces = "";
   my $ns;
   foreach $ns (@{$self->{SERVER}{namespaces}}) {
     $namespaces .= " ".$ns->GetStream();
+    $stream .= " ".$ns->GetStream();
   }
+  $stream .= ">";
 
   #---------------------------------------------------------------------------
-  # Now let's send the opening handshake.
+  # Then we send the opening handshake.
   #---------------------------------------------------------------------------
-  $self->{SERVER}{sock}->print(<<EOF) || return undef;
-<?xml version="1.0"?>
-<stream:stream xmlns:stream="http://etherx.jabber.org/streams" to="$self->{SERVER}{name}" xmlns="$self->{SERVER}{namespace}"$namespaces>
-EOF
+  $self->Send($stream) || return;
 
   #---------------------------------------------------------------------------
   # Create the XML::Parser and register our callbacks
@@ -237,23 +333,32 @@ EOF
   # Before going on let's make sure that the server responded with a valid
   # root tag and that the stream is open.
   #---------------------------------------------------------------------------
-  my ($buff, $timeout);
+  my $buff;
+  my $timeStart = time();
   while($self->{STATUS} == 0) {
-    if($self->{SERVER}{select}->can_read(1)) {
-      $self->{SERVER}{sock}->sysread($buff,1024);
+    if ($self->{SERVER}{select}->can_read(0)) {
+      $buff = $self->Read();
       $self->{SERVER}{parser}->parse_more($buff);
+      if ($STREAMERROR ne "") {
+	$self->SetErrorCode($STREAMERROR);
+	return;
+      }	
+
       # ToDo: we need to try/catch expat parsing errors here, no?
     } else {
-      $timeout++;
-      return undef if($timeout > $self->{SERVER}{timeout});
+      if ($timeout ne "") {
+	$timeout -= (time() - $timeStart);
+	if ($timeout <= 0) {
+	  $self->SetErrorCode("Timeout limit reached");
+	  return;
+	}
+      }
     }
     
-    return undef if($self->{SERVER}{select}->has_error(0));
+    return if($self->{SERVER}{select}->has_error(0));
   }
-  if($self->{STATUS} != 1) {
-    return undef;
-  }
-  return 1;
+  return if($self->{STATUS} != 1);
+  return $self->GetRoot();
 }
 
 
@@ -283,40 +388,60 @@ sub Disconnect {
 sub Process {
   my $self = shift;
   my($timeout) = @_;
-  my($buff);
+  $timeout = "" if !defined($timeout);
 
+  my($buff);
   
   #---------------------------------------------------------------------------
   # We need to keep track of what's going on in the function and tell the
-  # outside world about it:
-  #     0    connection open but no data.
+  # outside world about it so let's return something useful:
+  #     0    connection open but no data received.
   #     1    connection open and data received.
-  #   undef  connection closed
-  #   array  data that has been collected over time
+  #   undef  connection closed and error
+  #   array  connection open and the data that has been collected 
+  #          over time (No CallBack specified)
   #---------------------------------------------------------------------------
   my ($status) = 0;
   
   #---------------------------------------------------------------------------
   # Make sure the connection is active.
   #---------------------------------------------------------------------------
-  return undef unless($self->{STATUS} == 1);
+  return unless ($self->{STATUS} == 1);
   
   #---------------------------------------------------------------------------
-  # Use the proper timeout, either the one defined here or in the default.
-  # Either block until there is data, or wait a certain period of time and
-  # then return control to the user.
+  # Either block until there is data and we have parsed it all, or wait a 
+  # certain period of time and then return control to the user.
   #---------------------------------------------------------------------------
-  if($self->{SERVER}{select}->can_read($timeout eq "" ? $self->{SERVER}{timeout} : $timeout)) {
-    $status = 1;
-    $self->{SERVER}{sock}->sysread($buff,1024);
-    $self->{SERVER}{parser}->parse_more($buff);
-    return undef unless($self->{STATUS} == 1);
+  my $block = 1;
+  my $timeStart = time();
+  while($block == 1) {
+    if($self->{SERVER}{select}->can_read(0)) {
+      while($self->{SERVER}{select}->can_read(0)) {
+	$status = 1;
+	$self->{STATUS} = -1 if (!defined($buff = $self->Read()));
+	$self->{SERVER}{parser}->parse_more($buff);
+	if ($STREAMERROR ne "") {
+	  $self->SetErrorCode($STREAMERROR);
+	  return;
+	}
+	return unless($self->{STATUS} == 1);
+      }
+      $block = 0;
+    }
+
+    if ($timeout ne "") {
+      my $time = time;
+      $timeout -= ($time - $timeStart);
+      $timeStart = $time;
+      $block = 0 if ($timeout <= 0);
+    }
+    select(undef,undef,undef,.25);
   }
   
   #---------------------------------------------------------------------------
   # If the Select has an error then shut this party down.
   #---------------------------------------------------------------------------
-  return undef if($self->{SERVER}{select}->has_error(0));
+  return if($self->{SERVER}{select}->has_error(0));
   
   #---------------------------------------------------------------------------
   # If there are XML::Parser::Tree objects that have not been collected return
@@ -377,7 +502,50 @@ sub GetSock {
 ##############################################################################
 sub Send {
   my $self = shift;
-  $self->{SERVER}{sock}->print(@_);
+  $self->debug("XML::Stream: Send: (@_)");
+  $self->{SERVER}{sock}->print(@_) || return;
+  return 1;
+}
+
+
+##############################################################################
+#
+# Read - Takes the data from the server and returns a string
+#
+##############################################################################
+sub Read {
+  my $self = shift;
+  my $buff;
+  my $status = $self->{SERVER}{sock}->sysread($buff,1024);
+  $self->debug("XML::Stream: Read: ($buff)");
+  return $buff unless $status == 0;
+  $self->debug("XML::Stream: Read: ERROR");
+  return;
+}
+
+
+##############################################################################
+#
+# GetErrorCode - if you are returned an undef, you can call this function
+#                and hopefully learn more information about the problem.
+#
+##############################################################################
+sub GetErrorCode {
+  my $self = shift;
+  return (($self->{ERRORCODE} ne "") ? $self->{ERRORCODE} : $!);
+}
+
+
+##############################################################################
+#
+# SetErrorCode - sets the error code so that the caller can find out more
+#                information about the problem
+#
+##############################################################################
+sub SetErrorCode {
+  my $self = shift;
+  my ($errorcode) = @_;
+  $self->{ERRORCODE} = $errorcode;
 }
 
 
@@ -397,7 +565,7 @@ sub _handle_root {
   # Make sure we are receiving a valid stream on the same namespace.
   #---------------------------------------------------------------------------
   $self->{STATUS} = 
-    (($tag eq "stream:stream") && 
+    (($tag eq "stream:stream") && exists($att{'xmlns'}) &&
      ($att{'xmlns'} eq $self->{SERVER}{namespace})) ? 1 : -1;
 
   
@@ -406,7 +574,7 @@ sub _handle_root {
   # you'll need to check the namespace or the from attributes sent by the 
   # server.
   #---------------------------------------------------------------------------
-  $self->{ROOT} = %att;
+  $self->{ROOT} = \%att;
 
   #---------------------------------------------------------------------------
   # Now that we have gotten a root tag, let's look for the tags that make up
@@ -427,6 +595,7 @@ sub _handle_root {
 sub _handle_element {
   my $self = shift;
   my ($expat, $tag, %att) = @_;
+
   my @NEW;
   if($#{$self->{TREE}} < 0) {
     push @{$self->{TREE}}, $tag;
@@ -472,8 +641,13 @@ sub _handle_close {
   
   if($#{$self->{TREE}} < 1) {
     push @{$self->{TREE}}, $CLOSED;
-    &{$self->{NODE}}(@{$self->{TREE}});
-    $self->{TREE} = [];
+
+    if($self->{TREE}->[0] eq "stream:error") {
+      $STREAMERROR = $self->{TREE}[1]->[2];
+    } else {
+      &{$self->{NODE}}(@{$self->{TREE}});
+      $self->{TREE} = [];
+    }
   } else {
     push @{$self->{TREE}[$#{$self->{TREE}}]}, $CLOSED;
   }
